unit Progs;

interface

uses
  Graphics, Windows, SysUtils, Classes, Forms, Contnrs, Math;

const
  MAX_PROG_COUNT = 99;
  MAX_HINT_LEN = 28;
  MAX_CMNT_LEN = 232;
  MAX_STEP_COUNT = 99;
  DEF_STEP_COUNT = 8;
  TRAILING_CHAR = 13;
  DEF_SRC_FILE_HEADER =
    '#ifndef _LEL_PROG_H_' + #13#10 +
    '#define _LEL_PROG_H_' + #13#10 +
    #13#10 +
    '#ifdef __AVR__' + #13#10 +
    ' #include <avr/io.h>' + #13#10 +
    ' #include <avr/pgmspace.h>' + #13#10 +
    '#elif defined(ESP8266)' + #13#10 +
    ' #include <pgmspace.h>' + #13#10 +
    '#else' + #13#10 +
    ' #define PROGMEM' + #13#10 +
    '#endif' + #13#10 + #13#10;
  DEF_SRC_FILE_FOOTER =  #13#10 + '#endif // _LEL_PROG_H_';
  FILE_SIGNATURE = '!LEL';

resourcestring
  rsDefProgName = ' %d';
  rsErrorFileOpen1 = ' 1   %s';
  rsErrorFileOpen2 = ' 2   %s';
  rsErrorFileOpen3 = '   %s';
  rsErrorFileCreate = '   %s';

  rsGenVersion = '// : %d; %.2d.%.2d.%.4d' + #13#10;
  rsGenMaxProgDefine = '#define MAX_PROG %d';
  rsGenMinProgDefine = '#define MAX_PROG 0';
  rsGenPrgLenDecl = 'const unsigned char progLength[%d] PROGMEM =';
  rsGenPrgStpDecl = 'const unsigned char progSteps[%d] PROGMEM =';

type
  TLelFileHeader = packed record  // 256 bytes
    Signature: array[0..3] of AnsiChar;
    Version: Byte;
    Reserved1: Byte;
    Day:   Byte;
    Month: Byte;
    Year:  Word;
    Count: Byte;
    Reserved2: array[0..10] of AnsiChar; // 
    Comment: String[MAX_CMNT_LEN]; // , , 
    Footer: Byte;
  end;

  TLelProgHeader = packed record  // 32 bytes
    Count: Byte;
    Reserved: Word;
    Hint: String[MAX_HINT_LEN]; // 
  end;

  TProg = class(TObject)
  private
    FSteps: array of Byte;
    FHint: AnsiString;
  protected
    function GetStepCount: Integer;
    function GetSteps(Index: Integer): Byte;
    procedure SetStepCount(const Value: Integer);
    procedure SetSteps(Index: Integer; const Value: Byte);
  public
    constructor Create; overload;
    constructor Create(AStepCount: Integer; AHint: AnsiString); overload;
    procedure Clear; virtual;
    procedure Delete(Index: Integer); virtual;
    procedure Insert(Index: Integer); virtual;
    procedure Exchange(Index1, Index2: Integer); virtual;
    //function Add(AObject: TObject): Integer;
    property StepCount: Integer read GetStepCount write SetStepCount;
    property Steps[Index: Integer]: Byte read GetSteps write SetSteps;
    property Hint: AnsiString read FHint write FHint;
  end;

  TProgList = class(TObjectList)
  private
    FPointsPerMm: Byte;
    FEMF: TMetafile;
    FFileName: String;
    FSrcFileHeader: TStrings;
    FSrcFileFooter: TStrings;
    FSrcFilePrgLenDecl: String;
    FSrcFilePrgStpDecl: String;
    procedure SetSrcFileHeader(const Value: TStrings);
    procedure SetSrcFileFooter(const Value: TStrings);
    function GetFileName: String;
  protected
    function MmsToPixels(Value: Double): Integer;
    function GetProg(Index: Integer): TProg;
    procedure SetProg(Index: Integer; const Value: TProg);
  public
    constructor Create; virtual;
    destructor Destroy; override;
    function Add(AProg: TProg): Integer;
    procedure Insert(Index: Integer; AProg: TProg);
    procedure SaveToFile(const FileName: String); virtual;
    function LoadFromFile(const FileName: String): Boolean; virtual;
    procedure GenSrc(Items: TStrings); virtual;
    procedure Draw(ToScreen: Boolean = True); virtual;
    property Picture: TMetafile read FEMF;
    property FileName: String read GetFileName;
    property Progs[Index: Integer]: TProg read GetProg write SetProg; default;
    property SrcFileHeader: TStrings read FSrcFileHeader write SetSrcFileHeader;
    property SrcFileFooter: TStrings read FSrcFileFooter write SetSrcFileFooter;
    property SrcFilePrgLenDecl: String read FSrcFilePrgLenDecl write FSrcFilePrgLenDecl;
    property SrcFilePrgStpDecl: String read FSrcFilePrgStpDecl write FSrcFilePrgStpDecl;
  end;

  function TstBit(Value: Byte; Bit: Byte): Boolean;
  function SetBit(Value: Byte; Bit: Byte): Byte;
  function RstBit(Value: Byte; Bit: Byte): Byte;
  function InvBit(Value: Byte; Bit: Byte): Byte;

var
  LelFileHeader: TLelFileHeader;

implementation

function TstBit(Value: Byte; Bit: Byte): Boolean;
begin
  Result:= Value and (1 shl Bit) <> 0;
end;

function SetBit(Value: Byte; Bit: Byte): Byte;
begin
  Result:= Value or (1 shl Bit);
end;

function RstBit(Value: Byte; Bit: Byte): Byte;
begin
  Result:= Value and not (1 shl Bit);
end;

function InvBit(Value: Byte; Bit: Byte): Byte;
begin
  Result:= Value xor (1 shl Bit);
end;

{ TProg }

procedure TProg.Clear;
begin
  SetLength(FSteps, 0);
  //FillChar(FSteps[0], Length(FSteps), 0);
end;

constructor TProg.Create;
begin
  inherited Create;
  Clear;
end;

constructor TProg.Create(AStepCount: Integer; AHint: AnsiString);
begin
  inherited Create;
  SetLength(FSteps, AStepCount);
  FillChar(FSteps[0], Length(FSteps), 0);
  FHint:= AHint;
end;

procedure TProg.Delete(Index: Integer);
var
  Last: Integer;
begin
  Last:= High(FSteps);
  if Index < Last then
    Move(FSteps[Index + 1], FSteps[Index], (Last - Index) * SizeOf(Byte));
  SetLength(FSteps, Last);
end;

procedure TProg.Exchange(Index1, Index2: Integer);
var
  B: Byte;
begin
  B:= FSteps[Index1];
  FSteps[Index1]:= FSteps[Index2];
  FSteps[Index2]:= B;
end;

function TProg.GetStepCount: Integer;
begin
  Result:= Length(FSteps);
end;

function TProg.GetSteps(Index: Integer): Byte;
begin
  Result:= FSteps[Index];
end;

procedure TProg.Insert(Index: Integer);
var
  Len: Integer;
begin
   Len:= Length(FSteps);
   if Index >= Len then
     Index:= Len + 1;
   SetLength(FSteps, Len + 1);
   Move(FSteps[Index], FSteps[Index + 1], (Len - Index) * SizeOf(Byte));
   FSteps[Index]:= 0;
end;

procedure TProg.SetStepCount(const Value: Integer);
var
  L: Integer;
begin
  L:= Length(FSteps);
  SetLength(FSteps, Value);
  if Value > L then
    FillChar(FSteps[L], Length(FSteps) - L, 0);
end;

procedure TProg.SetSteps(Index: Integer; const Value: Byte);
begin
  FSteps[Index]:= Value;
end;

{ TProgList }

function TProgList.Add(AProg: TProg): Integer;
begin
  Result:= inherited Add(AProg);
end;

constructor TProgList.Create;
begin
  inherited Create;
  OwnsObjects:= True;
  FFileName:= '';
  Add(TProg.Create(DEF_STEP_COUNT, Format(rsDefProgName, [Count + 1])));
  FSrcFileHeader:= TStringList.Create;
  FSrcFileHeader.Text:= DEF_SRC_FILE_HEADER;
  FSrcFileFooter:= TStringList.Create;
  FSrcFileFooter.Text:= DEF_SRC_FILE_FOOTER;
  FSrcFilePrgLenDecl:= rsGenPrgLenDecl;
  FSrcFilePrgStpDecl:= rsGenPrgStpDecl;
  FPointsPerMm:= 24;
  FEMF:= TMetafile.Create;
  FEMF.Enhanced:= True;
end;

destructor TProgList.Destroy;
begin
  FSrcFileHeader.Free;
  inherited;
end;

const
  WW = 4.0;
  HH = 0.75;

procedure TProgList.Draw(ToScreen: Boolean);
var
  ACanvas: TMetafileCanvas;
  i, j, k, y, w, h, MaxStpCnt: Integer;
  r: TRect;
  s: String;
  B: Byte;
  OldFont: TFont;
begin
  MaxStpCnt:= 0;
  for i:= 0 to Count - 1 do
    MaxStpCnt:= Max(MaxStpCnt, Progs[i].StepCount);
  w:= (MaxStpCnt + 2) * MmsToPixels(WW);
  h:= (Count + 2) * MmsToPixels(HH);
  FEMF.Clear;
  FEMF.Width:= w;
  FEMF.Height:= h;
  ACanvas:= TMetafileCanvas.Create(FEMF, 0);
  OldFont:= TFont.Create;
  try
    FEMF.Width:= w;
    FEMF.Height:= h;
    // 
    ACanvas.Brush.Color:= clWhite;
    r:= Rect(0, 0, FEMF.Width, FEMF.Height);
    if ToScreen then
      ACanvas.Rectangle(r)
    else
      ACanvas.FillRect(r);
    SetBkMode(ACanvas.Handle, TRANSPARENT);
    //s:= '0123456789'; DrawText(ACanvas.Handle, PChar(s), Length(s), r, DT_SINGLELINE or DT_LEFT);

    OldFont.Assign(ACanvas.Font);
    ACanvas.Font.Name:= 'Courier New';
    ACanvas.Font.Size:= 10;
    ACanvas.Font.Color:= clRed;
    y:= MmsToPixels(HH);
    for i:= 0 to Count - 1 do
    begin
      ACanvas.Font.Color:= clBlack;
      r:= Rect(0, y, MmsToPixels(WW), y + MmsToPixels(HH));
      s:= Format('%d', [i]);
      DrawText(ACanvas.Handle, PChar(s), Length(s), r, DT_SINGLELINE or DT_VCENTER or DT_CENTER);
      for j:= 0 to Progs[i].StepCount - 1 do
      begin
        r:= Rect((j + 1) * MmsToPixels(WW), y, (j + 2) * MmsToPixels(WW), y + MmsToPixels(HH));
        InflateRect(r, -1, -1);
        ACanvas.Rectangle(r);
        s:= '';
        ACanvas.Font.Color:= clBlack;
        for k:= 1 to 4 do
        begin
          B:= Progs[i].Steps[j];
          if TstBit(B, 8 - k) then   // 
          begin
            if TstBit(B, 4 - k) then // 
            begin
              s:= s + '<=';
              ACanvas.Font.Color:= clBlue;
            end else
            begin
              s:= s + '=>';
              ACanvas.Font.Color:= clRed;
            end;
          end else
          begin
            s:= s + '==';
            //ACanvas.Font.Color:= clBlack;
          end;
          s:= s + ' ';
        end;

        //s:= Format('%d,%d', [i, j]);
        s:= Trim(s);
        InflateRect(r, 1, 1);
        SetBkMode(ACanvas.Handle, TRANSPARENT);
        DrawText(ACanvas.Handle, PChar(s), Length(s), r, DT_SINGLELINE or DT_VCENTER or DT_CENTER);
      end;
      y:= y + MmsToPixels(HH);
    end;
    ACanvas.Font.Assign(OldFont);
  finally
    OldFont.Free;
    ACanvas.Free;
  end;
end;

procedure TProgList.GenSrc(Items: TStrings);
var
  i, j, N: Integer;
  S: AnsiString;
begin
  Items.BeginUpdate;
  try
    Items.Clear;
    Items.AddStrings(FSrcFileHeader);
    Items.Add(Format('//    "%s"', [Application.Title]));
    Items.Add('// ' + LelFileHeader.Comment);
    with LelFileHeader do
      Items.Add(Format(rsGenVersion, [Version, Day, Month, Year]));
    Items.Add('// ./.  () ');
    Items.Add(Format(rsGenMaxProgDefine, [Count - 1]));
    Items.Add(rsGenMinProgDefine);
    Items.Add('');
    //       .    
    Items.Add(Format(FSrcFilePrgLenDecl + ' //  ', [Count]));
    Items.Add('{');
    S:= '';
    N:= 0;
    for i:= 0 to Count - 1 do
    begin
      if i = 0 then
        S:= IntToStr(Progs[i].StepCount)
      else
        S:= S + ',' + IntToStr(Progs[i].StepCount);
      N:= N + Progs[i].StepCount;
    end;
    Items.Add('  ' + S);
    Items.Add('};');
    Items.Add('');
    //   
    Items.Add(Format(FSrcFilePrgStpDecl + ' //  ', [N]));
    Items.Add('{');
    for i:= 0 to Count - 1 do
    begin
      S:= '';
      for j:= 0 to Progs[i].StepCount - 1 do
      begin
        if j = 0 then
          S:= '0x' + IntToHex(Progs[i].Steps[j], 2)
        else
          S:= S + ',' + '0x' + IntToHex(Progs[i].Steps[j], 2);
      end;
      if i <> (Count - 1) then
        S:= S + ',';
      Items.Add('  ' + S + '  // ' + Progs[i].Hint);
    end;
    Items.Add('};');
    Items.AddStrings(FSrcFileFooter);
  finally
    Items.EndUpdate;
  end;
end;

function TProgList.GetFileName: String;
begin
  Result:= FFileName;
end;

function TProgList.GetProg(Index: Integer): TProg;
begin
  Result:= inherited GetItem(Index) as TProg;
end;

procedure TProgList.Insert(Index: Integer; AProg: TProg);
begin
  inherited Insert(Index, AProg);
end;

function TProgList.LoadFromFile(const FileName: String): Boolean;
var
  fHandle, i, j: Integer;
  pHeader: TLelProgHeader;
  B: Byte;
  AProg: TProg;
begin
  Result:= False;
  fHandle:= FileOpen(FileName, fmOpenRead);
  if fHandle > 0 then
  begin
    FillChar(LelFileHeader, SizeOf(TLelFileHeader), 0);
    FileRead(fHandle, LelFileHeader, SizeOf(TLelFileHeader));
    //      
    if (LelFileHeader.Signature <> FILE_SIGNATURE) or (LelFileHeader.Footer <> TRAILING_CHAR) then
    begin
      Application.MessageBox(PChar(Format(rsErrorFileOpen1, [FileName])),
        PChar(Application.Title), MB_ICONWARNING or MB_OK);
      FileClose(fHandle);
      Exit;  
    end;
    Clear;
    for i:= 0 to LelFileHeader.Count - 1 do
    begin
      FillChar(pHeader, SizeOf(TLelProgHeader), 0);
      FileRead(fHandle, pHeader, SizeOf(TLelProgHeader));
      AProg:= TProg.Create;
      AProg.StepCount:= pHeader.Count;
      AProg.Hint:= pHeader.Hint;
      for j:= 0 to AProg.StepCount - 1 do
      begin
        FileRead(fHandle, B, SizeOf(B));
        AProg.Steps[j]:= B;
      end;
      FileRead(fHandle, B, SizeOf(B)); //    
      if B <> TRAILING_CHAR then
      begin
        Application.MessageBox(PChar(Format(rsErrorFileOpen2, [FileName])),
          PChar(Application.Title), MB_ICONWARNING or MB_OK);
        FileClose(fHandle);
        Exit;
      end;
      Add(AProg);
    end;
    FFileName:= FileName;
    FileClose(fHandle);
    Result:= True;
  end else
    raise Exception.CreateFmt(rsErrorFileOpen3, [FileName]);
end;

function TProgList.MmsToPixels(Value: Double): Integer;
begin
  Result:= Round(Value * FPointsPerMm);
end;

procedure TProgList.SaveToFile(const FileName: String);
var
  fHandle, i, j: Integer;
  pHeader: TLelProgHeader;
  B: Byte;
begin
  fHandle:= FileCreate(FileName);
  if fHandle > 0 then
  begin
    LelFileHeader.Count:= Count;
    LelFileHeader.Footer:= TRAILING_CHAR; //    
    FileWrite(fHandle, LelFileHeader, SizeOf(TLelFileHeader));
    for i:= 0 to Count - 1 do
    begin
      FillChar(pHeader, SizeOf(TLelProgHeader), 0);
      with pHeader do
      begin
        Count:= Progs[i].StepCount;
        Hint:= Progs[i].Hint;
      end;
      FileWrite(fHandle, pHeader, SizeOf(TLelProgHeader));
      for j:= 0 to Progs[i].StepCount - 1 do
      begin
        B:= Progs[i].Steps[j];
        FileWrite(fHandle, B, SizeOf(B));
      end;
      B:= TRAILING_CHAR; //    
      FileWrite(fHandle, B, SizeOf(B));
    end;
    FFileName:= FFileName;
    FileClose(fHandle);
  end else
    raise Exception.CreateFmt(rsErrorFileCreate, [FileName]);
end;

procedure TProgList.SetProg(Index: Integer; const Value: TProg);
begin
  inherited SetItem(Index, Value);
end;

procedure TProgList.SetSrcFileFooter(const Value: TStrings);
begin
  FSrcFileFooter.Assign(Value);
end;

procedure TProgList.SetSrcFileHeader(const Value: TStrings);
begin
  FSrcFileHeader.Assign(Value);
end;

end.
